print("import: 'mizani'")
import mizani

print("import: 'mizani.external'")
import mizani.external

print("import: 'mizani.tests'")
import mizani.tests

print("import: 'mizani.colors'")
import mizani.colors

