"""
Standard colors found in regular Crayola assortments from 1903 to
the present

Source: https://en.wikipedia.org/wiki/List_of_Crayola_crayon_colors
"""

crayon_rgb = {
    'almond': '#eed9c4',
    'antique brass': '#c88a65',
    'apricot': '#fdd5b1',
    'aquamarine': '#95e0e8',
    'asparagus': '#7ba05b',
    'banana mania': '#fbe7b2',
    'beaver': '#926f5b',
    'bittersweet': '#fe6f5e',
    'black': '#000000',
    'blue (i)': '#4997d0',
    'blue (ii)': '#4570e6',
    'blue (iii)': '#0066ff',
    'blue bell': '#9999cc',
    'blue-gray': '#c8c8cd',
    'blue-green': '#0095b7',
    'blue-violet': '#6456b7',
    'blush': '#db5079',
    'brick red': '#c62d42',
    'brilliant rose': '#e667ce',
    'brown': '#af593e',
    'burnt orange': '#ff7f49',
    'burnt sienna': '#e97451',
    'burnt umber': '#805533',
    'cadet blue': '#a9b2c3',
    'canary': '#ffff99',
    'caribbean green': '#00cc99',
    'carmine': '#e62e6b',
    'carnation pink': '#ffa6c9',
    'celestial blue': '#7070cc',
    'cerise': '#da3287',
    'cerulean': '#02a4d3',
    'cerulean blue': '#339acc',
    'charcoal gray': '#736a62',
    'cobalt blue': '#8c90c8',
    'copper': '#da8a67',
    'cornflower': '#93ccea',
    'cotton candy': '#ffb7d5',
    'dandelion': '#fed85d',
    'dark venetian red': '#b33b24',
    'denim': '#1560bd',
    'desert sand': '#edc9af',
    'eggplant': '#614051',
    'english vermilion': '#cc474b',
    'fern': '#63b76c',
    'forest green': '#5fa777',
    'fuchsia': '#c154c1',
    'fuzzy wuzzy': '#87421f',
    'gold (i)': '#92926e',
    'gold (ii)': '#e6be8a',
    'goldenrod': '#fcd667',
    'granny smith apple': '#9de093',
    'gray': '#8b8680',
    'green': '#3aa655',
    'green-blue': '#2887c8',
    'green-yellow': '#f1e788',
    'inchworm': '#afe313',
    'indian red': '#b94e48',
    'indigo': '#4f69c6',
    'jazzberry jam': '#a50b5e',
    'jungle green': '#29ab87',
    'lavender (i)': '#bf8fcc',
    'lavender (ii)': '#fbaed2',
    'lemon yellow': '#ffff9f',
    'light blue': '#8fd8d8',
    'light chrome green': '#bee64b',
    'light venetian red': '#e6735c',
    'macaroni and cheese': '#ffb97b',
    'madder lake': '#cc3336',
    'magenta': '#f653a6',
    'mahogany': '#ca3435',
    'maize': '#f2c649',
    'manatee': '#8d90a1',
    'mango tango': '#e77200',
    'maroon': '#c32148',
    'mauvelous': '#f091a9',
    'maximum blue': '#47abcc',
    'maximum blue green': '#30bfbf',
    'maximum blue purple': '#acace6',
    'maximum green': '#5e8c31',
    'maximum green yellow': '#d9e650',
    'maximum purple': '#733380',
    'maximum red': '#d92121',
    'maximum red purple': '#a63a79',
    'maximum yellow': '#fafa37',
    'maximum yellow red': '#f2ba49',
    'medium chrome green': '#6ca67c',
    'medium rose': '#d96cbe',
    'medium violet': '#8f47b3',
    'melon': '#febaad',
    'middle blue': '#7ed4e6',
    'middle blue green': '#8dd9cc',
    'middle blue purple': '#8b72be',
    'middle green': '#4d8c57',
    'middle green yellow': '#acbf60',
    'middle purple': '#d982b5',
    'middle red': '#e58e73',
    'middle red purple': '#a55353',
    'middle yellow': '#ffeb00',
    'middle yellow red': '#ecb176',
    'midnight blue': '#00468c',
    'mountain meadow': '#1ab385',
    'mulberry': '#c8509b',
    'navy blue': '#0066cc',
    'olive green': '#b5b35c',
    'orange': '#ff8833',
    'orange-red': '#ff5349',
    'orange-yellow': '#f8d568',
    'orchid': '#e29cd2',
    'outer space': '#2d383a',
    'pacific blue': '#009dc4',
    'peach': '#ffcba4',
    'periwinkle': '#c3cde6',
    'permanent geranium lake': '#e12c2c',
    'pig pink': '#fdd7e4',
    'pine green': '#01786f',
    'pink flamingo': '#fc74fd',
    'pink sherbert': '#f7a38e',
    'plum': '#8e3179',
    'purple heart': '#652dc1',
    "purple mountains' majesty": '#d6aedd',
    'raw sienna': '#d27d46',
    'raw sienna (i)': '#e6bc5c',
    'raw umber': '#665233',
    'razzmatazz': '#e30b5c',
    'red': '#ed0a3f',
    'red-orange': '#ff681f',
    'red-violet': '#bb3385',
    "robin's egg blue": '#00cccc',
    'royal purple': '#6b3fa0',
    'salmon': '#ff91a4',
    'scarlet': '#fd0e35',
    'sea green': '#93dfb8',
    'sepia': '#9e5b40',
    'shadow': '#837050',
    'shamrock': '#33cc99',
    'silver': '#c9c0bb',
    'sky blue': '#76d7ea',
    'spring green': '#ecebbd',
    'sunset orange': '#fe4c40',
    'tan': '#d99a6c',
    'teal blue': '#008080',
    'thistle': '#ebb0d7',
    'tickle me pink': '#fc80a5',
    'timberwolf': '#d9d6cf',
    'tropical rain forest': '#00755e',
    'tumbleweed': '#dea681',
    'turquoise blue': '#6cdae7',
    'ultramarine blue': '#3f26bf',
    'van dyke brown': '#664228',
    'venetian red': '#cc553d',
    'violet (i)': '#732e6c',
    'violet (ii)': '#8359a3',
    'violet-blue': '#766ec8',
    'violet-red': '#f7468a',
    'vivid tangerine': '#ff9980',
    'vivid violet': '#803790',
    'white': '#ffffff',
    'wild blue yonder': '#7a89b8',
    'wild strawberry': '#ff3399',
    'wisteria': '#c9a0dc',
    'yellow': '#fbe870',
    'yellow-green': '#c5e17a',
    'yellow-orange': '#ffae42',
}
